/*
 * tinycan.c - Tiny-CAN I-XL / II-XL / IV-XL interface driver (using tty line discipline)
 *
 * This file is derived from linux/drivers/net/slip.c and linux/drivers/net/can/slcan.c
 *
 * slip.c Authors  : Laurence Culhane <loz@holmes.demon.co.uk>
 *                   Fred N. van Kempen <waltje@uwalt.nl.mugnet.org>
 * slcan.c Author  : Oliver Hartkopp <socketcan@hartkopp.net>
 * tinycan.c Author: Klaus Demlehner <klaus@mhs-elektronik.de>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307. You can also get it
 * at http://www.gnu.org/licenses/gpl.html
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 * Send feedback to <klaus@mhs-elektronik.de>
 *
 */

#include <linux/module.h>
#include <linux/moduleparam.h>

#include <linux/err.h>
#include <linux/kernel.h>
#include <linux/kthread.h>
#include <linux/uaccess.h>
#include <linux/bitops.h>
#include <linux/string.h>
#include <linux/tty.h>
#include <linux/errno.h>
#include <linux/netdevice.h>
#include <linux/skbuff.h>
#include <linux/rtnetlink.h>
#include <linux/if_arp.h>
#include <linux/if_ether.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/tty_flip.h>
#include <linux/can.h>
#include <linux/can/dev.h>
#include <linux/can/error.h>
#include <linux/usb.h>
#include <linux/usb/serial.h>


#define DRV_NAME "tinycan"
static __initdata const char banner[] =
	KERN_INFO "tinycan: CAN driver for Tiny-CAN interfaces\n";

MODULE_ALIAS_LDISC(N_SLCAN); // <*> Wurde von slcan übernommen
MODULE_DESCRIPTION("Tiny-CAN interface driver");
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Klaus Demlehner <klaus@mhs-elektronik>");
MODULE_VERSION("1.1.1");

#define TCAN_MAGIC 0x53CA  // <*> Der Wert wurde von slcan übernommen:-) ???

static unsigned char CRCTabelle[] = {
    0x00, 0x5e, 0xbc, 0xe2, 0x61, 0x3f, 0xdd, 0x83,
    0xc2, 0x9c, 0x7e, 0x20, 0xa3, 0xfd, 0x1f, 0x41,
    0x9d, 0xc3, 0x21, 0x7f, 0xfc, 0xa2, 0x40, 0x1e,   // 10
    0x5f, 0x01, 0xe3, 0xbd, 0x3e, 0x60, 0x82, 0xdc,
    0x23, 0x7d, 0x9f, 0xc1, 0x42, 0x1c, 0xfe, 0xa0,   // 20
    0xe1, 0xbf, 0x5d, 0x03, 0x80, 0xde, 0x3c, 0x62,
    0xbe, 0xe0, 0x02, 0x5c, 0xdf, 0x81, 0x63, 0x3d,   // 30
    0x7c, 0x22, 0xc0, 0x9e, 0x1d, 0x43, 0xa1, 0xff,
    0x46, 0x18, 0xfa, 0xa4, 0x27, 0x79, 0x9b, 0xc5,   // 40
    0x84, 0xda, 0x38, 0x66, 0xe5, 0xbb, 0x59, 0x07,
    0xdb, 0x85, 0x67, 0x39, 0xba, 0xe4, 0x06, 0x58,   // 50
    0x19, 0x47, 0xa5, 0xfb, 0x78, 0x26, 0xc4, 0x9a,
    0x65, 0x3b, 0xd9, 0x87, 0x04, 0x5a, 0xb8, 0xe6,   // 60
    0xa7, 0xf9, 0x1b, 0x45, 0xc6, 0x98, 0x7a, 0x24,
    0xf8, 0xa6, 0x44, 0x1a, 0x99, 0xc7, 0x25, 0x7b,   // 70
    0x3a, 0x64, 0x86, 0xd8, 0x5b, 0x05, 0xe7, 0xb9,
    0x8c, 0xd2, 0x30, 0x6e, 0xed, 0xb3, 0x51, 0x0f,   // 80
    0x4e, 0x10, 0xf2, 0xac, 0x2f, 0x71, 0x93, 0xcd,
    0x11, 0x4f, 0xad, 0xf3, 0x70, 0x2e, 0xcc, 0x92,   // 90
    0xd3, 0x8d, 0x6f, 0x31, 0xb2, 0xec, 0x0e, 0x50,
    0xaf, 0xf1, 0x13, 0x4d, 0xce, 0x90, 0x72, 0x2c,   // A0
    0x6d, 0x33, 0xd1, 0x8f, 0x0c, 0x52, 0xb0, 0xee,
    0x32, 0x6c, 0x8e, 0xd0, 0x53, 0x0d, 0xef, 0xb1,   // B0
    0xf0, 0xae, 0x4c, 0x12, 0x91, 0xcf, 0x2d, 0x73,
    0xca, 0x94, 0x76, 0x28, 0xab, 0xf5, 0x17, 0x49,   // C0
    0x08, 0x56, 0xb4, 0xea, 0x69, 0x37, 0xd5, 0x8b,
    0x57, 0x09, 0xeb, 0xb5, 0x36, 0x68, 0x8a, 0xd4,   // D0
    0x95, 0xcb, 0x29, 0x77, 0xf4, 0xaa, 0x48, 0x16,
    0xe9, 0xb7, 0x55, 0x0b, 0x88, 0xd6, 0x34, 0x6a,   // E0
    0x2b, 0x75, 0x97, 0xc9, 0x4a, 0x14, 0xf6, 0xa8,
    0x74, 0x2a, 0xc8, 0x96, 0x15, 0x4b, 0xa9, 0xf7,   // F0
    0xb6, 0xe8, 0x0a, 0x54, 0xd7, 0x89, 0x6b, 0x35};


#define updcrc8(cp, crc) (CRCTabelle[crc ^ cp])

#define CMD_TYPE_EXT1 0x80
#define CMD_TYPE_EXT2 0xC0

#define CMD_PING                   0x7E
#define CMD_CAN0_SET               0x01 | CMD_TYPE_EXT1
#define CMD_CAN0_SET_MODE          0x03 | CMD_TYPE_EXT1
#define CMD_COM                    0x24 | CMD_TYPE_EXT1
#define CMD_SET_TIMESTAMP          0x25 | CMD_TYPE_EXT1
#define CMD_GET_STATUS_ALL         0x41

#define CAN_SPEED_INDEX         0
#define CAN_SPEED_BTR           1
#define CAN_TX_ACK_ENABLE       2


#define TAR_STREAM_STATUS     0x04
#define TAR_STREAM_RXD        0x40

#define TAR_CAN_RXD_PEND   0x08
#define TAR_CAN_RXD_OV     0x10

#define TAR_CAN_TXD_MASK   0x60
#define TAR_CAN_TXD_EMPTY  0x00
#define TAR_CAN_TXD_PEND   0x20
#define TAR_CAN_TXD_LIMIT  0x40
#define TAR_CAN_TXD_OV     0x60

#define TCAN_MTU 2048
#define HW_TX_FIFO_LIMIT 36
#define HW_TX_FIFO_SIZE  72
#define TX_CAN_TX_LIMIT 9
#define STATUS_READ_TIMEOUT 1000  // 1000 ms
#define RXD_RECEIVE_TIMEOUT 50    // 50 ms

#define RX_LOCK 0
#define RX_WAIT 1
#define RX_PCNT 2
#define RX_DATA 3
#define RX_CRC  4
#define RX_FIN  5

#define ERR_SM_NACK      -1
#define ERR_SM_WRONG_ACK -2
#define ERR_SM_WRONG_CRC -3
#define ERR_SM_TIMEOUT   -4
#define ERR_PARAM_OV     -5

#define ERR_DEVICE_POWER_UP       -10
#define ERR_CAN_STATUS            -11
#define ERR_UNSUPPORTED_CAN_SPEED -12
#define ERR_CMD_CAN_SET           -13
#define ERR_CMD_SET_TIMESTAMP     -14
#define ERR_CMD_CAN_SET_MODE      -15
#define ERR_CMD_GET_STATUS        -16

/* CAN related stuff */
#define CANMSG_BUF_SIZE		0x100	/* 256 messages */

#define CANMSG_BUF_EMPTY	0x00	/* fifo is empty */
#define CANMSG_BUF_PEND		0x01	/* message(s) pending */
#define CANMSG_BUF_FULL		0x02	/* fifo is full */
#define CANMSG_BUF_OVR		0x04	/* fifo overrun */

#define TX_ECHO_SKB_MAX  1

#define HW_TCAN_ERROR 0
#define HW_TCAN_I_XL  1
#define HW_TCAN_II_XL 2
#define HW_TCAN_IV_XL 3

#define MHS_CAN_HW_OFFLINE    0
#define MHS_CAN_BUS_DO_INIT   1
#define MHS_CAN_BUS_INIT      2
#define MHS_CAN_BUS_DO_STOP   3
#define MHS_CAN_BUS_STOP      4
#define MHS_CAN_BUS_DO_START  5
#define MHS_CAN_BUS_RUN       6




struct canmsg_fifo 
  {
  volatile unsigned long head;
  volatile unsigned long tail;
  volatile unsigned long status;
  struct can_frame msg[CANMSG_BUF_SIZE];
  };


struct tcandrv_priv 
  {
  struct can_priv can; /* must be the first member */
  struct net_device *dev;

  int magic;
  /* Various fields. */
  struct tty_struct *tty;		// ptr to TTY structure	     
  volatile unsigned long mhs_can_mode;
  unsigned long tiny_can_hardware;
  unsigned long enable_timestamp;
  spinlock_t fifo_lock;
  spinlock_t lock;
  struct task_struct *tcan_thread;

  unsigned long status_timeout;
  int read_status_flag;
  struct can_berr_counter bec;
  unsigned char can_status_old;
  unsigned char can_rx_ov_old;

  unsigned char tx_cmd;
  unsigned long tx_param_cnt;
  unsigned char tx_param[TCAN_MTU];  
  struct canmsg_fifo txfifo;          // transmit fifo
  uint32_t hw_txd_fifo_count;

  volatile int rx_state;
  unsigned char ack_byte;
  unsigned char rx_crc;
  unsigned long rx_error;
  unsigned long rx_idx;
  unsigned long rx_param_cnt;
  unsigned char rx_param[TCAN_MTU];
  
  unsigned char	xbuff[TCAN_MTU];	// transmitter buffer	          
  };


#define MHS_CAN_CORE_CLOCK 16000000

static struct can_bittiming_const mhs_usb_bittiming_const = 
  {
  .name = "Tiny-CAN XL",
  .tseg1_min = 2,
  .tseg1_max = 16,
  .tseg2_min = 1,
  .tseg2_max = 8,
  .sjw_max = 4,
  .brp_min = 1,
  .brp_max = 1024,
  .brp_inc = 1,
  };


static int ExecuteCmd(struct tcandrv_priv *tc)
{
unsigned char c, crc;
uint16_t i, param_cnt;
uint32_t tx_cnt;
unsigned char *tx_puf;

tc->rx_state = RX_WAIT;
// Daten PC -> Tiny-CAN
tx_puf = &tc->xbuff[0];
c = tc->tx_cmd;
// **** Comando senden
*tx_puf++ = c;
crc = CRCTabelle[c];
c &= 0xC0;                       // Commando Type -> Single, Ext I oder II
// **** Parameter länge senden
if (c == 0x80)                   // Ext. I
  {
  param_cnt = (uint16_t)tc->tx_param_cnt;
  tx_cnt = param_cnt + 3;
  c = (unsigned char)param_cnt;
  crc = updcrc8(c, crc);
  *tx_puf++ = c;
  }
else if (c == 0xC0)              // Ext. II
  {
  param_cnt = (uint16_t)tc->tx_param_cnt;
  tx_cnt = param_cnt + 4;
  c = (unsigned char)((param_cnt >> 8) & 0x3F);
  crc = updcrc8(c, crc);
  *tx_puf++ = c;
  c = (unsigned char)param_cnt;
  crc = updcrc8(c, crc);
  *tx_puf++ = c;
  }
else
  {
  param_cnt = 0;
  tx_cnt = 2;
  }
// **** Parameter senden
for (i = 0; i < param_cnt; i++)
  {
  c = tc->tx_param[i];
  crc = updcrc8(c, crc);
  *tx_puf++ = c;
  }
// **** CRC senden
crc = CRCTabelle[CRCTabelle[crc]];
*tx_puf = crc;
set_bit(TTY_DO_WRITE_WAKEUP, &tc->tty->flags);
tc->tty->ops->write(tc->tty, tc->xbuff, tx_cnt);
// **** Daten Tiny-CAN -> PC
(void)schedule_timeout_uninterruptible(msecs_to_jiffies(RXD_RECEIVE_TIMEOUT));
if (tc->rx_state == RX_FIN)
  return(tc->rx_error);
else
  return(ERR_SM_TIMEOUT);
}


/* Read error/status frames */
static void CanRxDProcessErrMessage(struct tcandrv_priv *tc, struct can_frame *cf)
{
struct sk_buff *skb;
struct net_device_stats *stats;
struct can_frame *err_cf;
int rx_errors, tx_errors;
unsigned char rxerr, txerr;

rx_errors = 0;
tx_errors = 0;
stats = &tc->dev->stats;
	
if (!(skb = alloc_can_err_skb(tc->dev, &err_cf)))
  return;

rxerr = cf->data[0]; // Receive Error Counter
txerr = cf->data[1]; // Transmit Error Counter

if (cf->can_id & 0x0080)  // Bus Off
  {
  tc->can.state = CAN_STATE_BUS_OFF;
  tc->can.can_stats.bus_off++;
  err_cf->can_id |= CAN_ERR_BUSOFF;
  can_bus_off(tc->dev);
  }    
else if (cf->can_id & 0x0020)  // Error Passiv
  {
  tc->can.state = CAN_STATE_ERROR_PASSIVE;
  tc->can.can_stats.error_passive++;
  err_cf->can_id |= CAN_ERR_CRTL;
  if (txerr || rxerr)
    err_cf->data[1] = (txerr > rxerr) ? CAN_ERR_CRTL_TX_PASSIVE : CAN_ERR_CRTL_RX_PASSIVE;
  else
    err_cf->data[1] = CAN_ERR_CRTL_TX_PASSIVE | CAN_ERR_CRTL_RX_PASSIVE;
  }  
else if (cf->can_id & 0x0040)  // Error Warning
  {
  tc->can.state = CAN_STATE_ERROR_WARNING;
  tc->can.can_stats.error_warning++;
  err_cf->can_id |= CAN_ERR_CRTL;
  err_cf->data[1] = (txerr > rxerr) ? CAN_ERR_CRTL_TX_WARNING : CAN_ERR_CRTL_RX_WARNING;
  } 
else
  {
  tc->can.state = CAN_STATE_ERROR_ACTIVE;
  tc->can.can_stats.bus_error++;
  err_cf->can_id |= CAN_ERR_PROT;
  err_cf->data[2] = CAN_ERR_PROT_ACTIVE;
  }   

switch (cf->can_id & 0x07)
  {
  case 1 : { // Stuff Error
           err_cf->data[2] |= CAN_ERR_PROT_STUFF;
           rx_errors = 1;
           break;
           }
  case 2 : { // Form Error
           err_cf->data[2] |= CAN_ERR_PROT_FORM;
           rx_errors = 1;
           break;
           }
  case 3 : { // Ack Error
           err_cf->can_id |= CAN_ERR_ACK;
           tx_errors = 1;
           break;
           }
  case 4 : { // Bit1 Error
           err_cf->data[2] |= CAN_ERR_PROT_BIT1;
           tx_errors = 1;
           break;
           }
  case 5 : { // Bit0 Error
           err_cf->data[2] |= CAN_ERR_PROT_BIT0;
           tx_errors = 1;
           break;
           }
  case 6 : { // CRC Error	
           err_cf->data[2] |= CAN_ERR_PROT_UNSPEC;
           err_cf->data[3] |= CAN_ERR_PROT_LOC_CRC_SEQ | CAN_ERR_PROT_LOC_CRC_DEL;
           rx_errors = 1;
           break;
           }
  }

if (tx_errors) 
  {
  err_cf->data[2] |= CAN_ERR_PROT_TX;
  stats->tx_errors++;
  }
if (rx_errors)
  stats->rx_errors++;
err_cf->data[6] = txerr;
err_cf->data[7] = rxerr;
tc->bec.txerr = txerr;
tc->bec.rxerr = rxerr;
netif_rx(skb);
stats->rx_packets++;
stats->rx_bytes += 8;
}


static int32_t CanRxDProcessMessages(struct tcandrv_priv *tc, unsigned char *data, int32_t size)
{
struct net_device_stats *stats;
struct sk_buff *skb;
struct can_frame cf;
struct can_frame *new_cf;
unsigned char header, len, tmr_len;
int txd_msg;
uint32_t ctrlmode, t;

ctrlmode = tc->can.ctrlmode;
stats = &tc->dev->stats;
if (tc->enable_timestamp)
  tmr_len = 4;
else
  tmr_len = 0;
while (size)
  {
  txd_msg = 0;
  header = *data++;  
  len = header & 0x0F;
  cf.can_dlc = len;              // Dlc extrahieren
  *(u64 *) (&cf.data) = 0;       // clear payload
  if (!(header & 0x80))
    {  // STD Frame
    size -= (2 + tmr_len + len);
    if (size < 0)
      return(-1);
    cf.can_id = *data++ | (((uint16_t)header << 4) & 0x0700); // Id extrahieren
    }
  else
    {
    if (header & 0x20)
      len = 0; // RTR Frame
    if (header & 0x10)
      txd_msg = 1;               // TxD Message
    if (header & 0x40)
      {  // EFF Frame
      size -= (5 + tmr_len + len);
      if (size < 0)
        return(-1);
      cf.can_id = (le32_to_cpu(*((uint32_t *)data))) | CAN_EFF_FLAG; // extended frame format
      data += 4;
      //cf.can_id |= CAN_EFF_FLAG; // extended frame format
      }
    else
      {  // STD Frame
      size -= (3 + tmr_len + len);
      if (size < 0)
        return(-1);
      cf.can_id = le16_to_cpu(*((uint16_t *)data));
      data += 2;
      }
    if (header & 0x20) 
      cf.can_id |= CAN_RTR_FLAG; // RTR frame
    }
  // **** Daten lesen
  if (len)
    {
    memcpy(&cf.data[0], data, len); // Daten in FIFO kopieren
    data += len;
    }    
  // **** Time Stamp lesen
  if (tc->enable_timestamp)
    {
    t = le32_to_cpu(*((uint32_t *)data)); // Zeitstempel lesen 0,1 ms Aufloesung  
    data += 4;
    //Sec = t / 10000;
    //USec = (t % 10000) * 100;
    }
  
  if (tc->tiny_can_hardware == HW_TCAN_IV_XL)
    {
    if ((txd_msg) && (!(cf.can_id & CAN_EFF_FLAG)) && (cf.can_id & 0x8000))   // Error Message
      {
      CanRxDProcessErrMessage(tc, &cf);
      continue;
      }
    }
  
  if (txd_msg)
    {
    stats->tx_packets++;
    stats->tx_bytes += len;
    if (!(ctrlmode & CAN_CTRLMODE_LOOPBACK))        
      continue;
    }
  else
    {
    stats->rx_packets++;
    stats->rx_bytes += len;
    }
  if (!(skb = alloc_can_skb(tc->dev, &new_cf)))
    return(-1);
   memcpy(new_cf, &cf, sizeof(struct can_frame));
   //skb->tstamp = <*>
  netif_rx(skb);
  }
return(1);
}


static int32_t CanTxDProcessMessage(struct tcandrv_priv *tc, unsigned char *dst, struct can_frame *cf)
{
unsigned char len; 
unsigned short std_id; 
  
len = cf->can_dlc;
if (cf->can_id & CAN_EFF_FLAG)
  {
  if (cf->can_id & CAN_RTR_FLAG)
    {
    *dst++ = len | 0xE0;  // EXT / RTR
    len = 0;
    }
  else
    *dst++ = len | 0xC0;  // EXT
  *(uint32_t *)dst = cpu_to_le32(cf->can_id & CAN_EFF_MASK);
  dst += 4;
  if (len)      
    memcpy(dst, &cf->data[0], len);               
  len += 5;
  }
else
  {
  std_id = cpu_to_le16(cf->can_id & CAN_SFF_MASK);
  if (cf->can_id & CAN_RTR_FLAG)
    {  // STD / RTR
    *dst++ = len | 0xA0;
    *(uint16_t *)dst = std_id;
    dst += 2;
    len = 3;
    }
  else
    {  // STD
    *dst++ = (unsigned char)(((std_id >> 4) & 0x70) | len);
    *dst++ = (unsigned char)std_id;
    memcpy(dst, &cf->data[0], len);     
    len += 2;
    }
  }
return(len);
}


static int32_t CanProcessStatus(struct tcandrv_priv *tc, unsigned char *src, int32_t size)
{
int32_t err;
struct sk_buff *skb;
struct can_frame *cf;
unsigned char status, can_status, can_rx_ov;

tc->status_timeout = jiffies + ((STATUS_READ_TIMEOUT * HZ) / 1000);
if (size < 1)
  return(-1);
err = 1;
status = *src;
can_status = status & 0x07;
if (status & TAR_CAN_RXD_OV)  // RxD overflow
  can_rx_ov = 1; 
else
  can_rx_ov = 0;
if (can_status == 0x07)                
  return(ERR_DEVICE_POWER_UP);                            
if ((can_status != tc->can_status_old) || (can_rx_ov != tc->can_rx_ov_old))
  {  
  tc->can_rx_ov_old = can_rx_ov;
  tc->can_status_old = can_status;
  if (((can_status >= 1) && (can_status <= 3)) || (can_rx_ov))
    {
    if (!(skb = alloc_can_err_skb(tc->dev, &cf)))    
      err = -1;    
    else
      {    
      switch (can_status)
        {
        case 1 : { // Error warning 
                 cf->can_id = CAN_ERR_CRTL;
                 cf->data[1] = CAN_ERR_CRTL_TX_WARNING | CAN_ERR_CRTL_RX_WARNING;
                 break;
                 }
        case 2 : { // Error passiv
                 cf->can_id = CAN_ERR_CRTL;
                 cf->data[1] = CAN_ERR_CRTL_TX_PASSIVE | CAN_ERR_CRTL_RX_PASSIVE;
                 break;
                 }
        case 3 : {
                 cf->can_id = CAN_ERR_BUSOFF;
                 can_bus_off(tc->dev);                     
                 break;
                 }     
        }    
      if (can_rx_ov)
        {
        cf->can_id |= CAN_ERR_CRTL;
        cf->data[1] |= CAN_ERR_CRTL_RX_OVERFLOW;
        }    
      netif_rx(skb);
      }
    }    
  }
if ((status & TAR_CAN_TXD_MASK) == TAR_CAN_TXD_EMPTY)
  tc->hw_txd_fifo_count = 0;
else if ((status & TAR_CAN_TXD_MASK)  == TAR_CAN_TXD_PEND)
  {
  if (tc->hw_txd_fifo_count > HW_TX_FIFO_LIMIT)
    tc->hw_txd_fifo_count = HW_TX_FIFO_LIMIT;
  }
return(err);
}


static int32_t MainCommProc(struct tcandrv_priv *tc)
{
int32_t size, err, res, rd_cnt, count;
unsigned char *data;
unsigned char type;

/**********************************/
/* Get Kommando Daten aufbereiten */
/**********************************/
type = 0;
size = 1;

// ****** Festlegen ob Status abgefragt wird
if ((tc->read_status_flag) || (time_after(jiffies, tc->status_timeout)))
  {
  tc->read_status_flag = 0;
  type |= TAR_STREAM_STATUS;  
  }
data = &tc->tx_param[1];
// ****** Versand von CAN Nachrichten
if ((tc->hw_txd_fifo_count < HW_TX_FIFO_LIMIT) && (tc->txfifo.status & CANMSG_BUF_PEND))
  {
  count = HW_TX_FIFO_LIMIT - tc->hw_txd_fifo_count;
  if (count > TX_CAN_TX_LIMIT)
    count = TX_CAN_TX_LIMIT;
  spin_lock(&tc->fifo_lock);
  rd_cnt = tc->txfifo.head < tc->txfifo.tail ?
           (tc->txfifo.tail - tc->txfifo.head) :
           (tc->txfifo.tail - tc->txfifo.head + CANMSG_BUF_SIZE);
  spin_unlock(&tc->fifo_lock);
  if (rd_cnt > count)
    rd_cnt = count;  
  while (rd_cnt--)
    {           
    count = CanTxDProcessMessage(tc, data, &tc->txfifo.msg[tc->txfifo.head]);
    size += count;
    data += count;
    tc->hw_txd_fifo_count++;

    spin_lock(&tc->fifo_lock);
    tc->txfifo.head = (++tc->txfifo.head) % (CANMSG_BUF_SIZE);
    if (tc->txfifo.head == tc->txfifo.tail) 
      {
      tc->txfifo.status = CANMSG_BUF_EMPTY;
      spin_unlock(&tc->fifo_lock);
      break;
      }
    tc->txfifo.status &= ~(CANMSG_BUF_FULL | CANMSG_BUF_OVR);
    spin_unlock(&tc->fifo_lock);
    }
  }

tc->tx_param[0] = type;
tc->tx_param_cnt = size;
tc->tx_cmd = CMD_COM; 
err = ExecuteCmd(tc);
if (!err)
  {
  /**********************************/
  /* Empfangene Daten verarbeiten   */
  /**********************************/
  if (tc->rx_param_cnt >= 2)
    {
    size = tc->rx_param_cnt-1;
    data = &tc->rx_param[0];
    type = *data++;
    // **** Status
    if (type & TAR_STREAM_STATUS)
      {      
      if ((res = CanProcessStatus(tc, data, size)) < 0)
        err = res;
      else
        {
        size -= res;
        data += res;
        }
      }
    // **** CAN Nachrichten Empfang
    if (!err)
      {
      if (type & TAR_STREAM_RXD)
        {
        if ((res = CanRxDProcessMessages(tc, data, size)) < 0)
          err = res;
        }
      }
    }
  }
return(err);
}


static int32_t StartStopTinyCan(struct tcandrv_priv *tc, int start)
{
struct can_bittiming *bt;
//unsigned char i;
uint32_t s, ctrlmode, brp;

ctrlmode = tc->can.ctrlmode;
// **** CAN Bus starten & Tx u. Rx Fifo/Overrun löschen
if (start)
  {
  bt = &tc->can.bittiming;
  brp = bt->brp - 1;
  s = (brp & 0x003F) |
      ((bt->sjw - 1) << 6) |
      ((bt->prop_seg + bt->phase_seg1 - 1) << 8) |
      ((bt->phase_seg2 - 1) << 12) |
      ((brp << 10) & 0x000F0000);
  netdev_info(tc->dev, "CAN-Baudrate:%d (BTR: 0x%08X)\n", bt->bitrate, s);

  tc->tx_cmd = CMD_CAN0_SET;
  tc->tx_param_cnt = 5;
  tc->tx_param[0] = CAN_SPEED_BTR;  // Index
  *(uint32_t *)(&tc->tx_param[1]) = cpu_to_le32(s);
  if (ExecuteCmd(tc))
    return(ERR_CMD_CAN_SET);

  tc->tx_param[0] = 0xF1;  // OP_CAN_START
  if (ctrlmode & CAN_CTRLMODE_LISTENONLY)
    tc->tx_param[0] = 0xF4;  // OP_CAN_START_LOM
  else
    {
    if (tc->tiny_can_hardware == HW_TCAN_IV_XL)
      {
      if (ctrlmode & CAN_CTRLMODE_ONE_SHOT)
        tc->tx_param[0] = 0xF5;  // OP_CAN_START_NO_RETRANS 
      }
    }
  tc->can_status_old = 0xFF;  // Status neu auswerten
  }
else
  tc->tx_param[0] = 0xF2;  // OP_CAN_STOP
tc->tx_cmd = CMD_CAN0_SET_MODE;
tc->tx_param_cnt = 1;
if (ExecuteCmd(tc))
  return(ERR_CMD_CAN_SET_MODE);
return(0);
}


int32_t TinyCanGetStatus(struct tcandrv_priv *tc, unsigned char *status)
{
tc->tx_cmd = CMD_GET_STATUS_ALL;
tc->tx_param_cnt = 0;
if (ExecuteCmd(tc))
  return(ERR_CMD_GET_STATUS);
else if (tc->rx_param_cnt < 1)
  return(ERR_CMD_GET_STATUS);
else
  *status = tc->rx_param[0];
return(0);
}


static int32_t InitTinyCan(struct tcandrv_priv *tc)
{
int32_t res;
uint32_t ctrlmode;

ctrlmode = tc->can.ctrlmode;
tc->enable_timestamp = 0;
if ((res = StartStopTinyCan(tc, 0)) < 0)  // Bus Stop
  return(res);
if (tc->tiny_can_hardware == HW_TCAN_IV_XL)
  {
  //tc->enable_timestamp = 1;   // <*> ändern!  
  // **** Disabel Hardware Timestamp
  tc->tx_cmd = CMD_SET_TIMESTAMP; 
                                  // Hardware Timestamps
  tc->tx_param_cnt = 1;
  tc->tx_param[0] = 0;  // 1 = Enable Hardware Timestamps / 0 = Disable <*> ändern!
  if (ExecuteCmd(tc))
    return(ERR_CMD_SET_TIMESTAMP);
  }
if (ctrlmode & CAN_CTRLMODE_LOOPBACK)
  {
  // **** Disable CAN Tx ACK
  tc->tx_cmd =  CMD_CAN0_SET;
  tc->tx_param_cnt = 2;
  tc->tx_param[0] = CAN_TX_ACK_ENABLE;
  tc->tx_param[1] = 1;  // 1 = Enable / 0 = Disable 
  if (ExecuteCmd(tc))
    return(ERR_CMD_CAN_SET);
  }
return(0);
}


static int tcan_thread_exec(void *arg)
{
struct tcandrv_priv *tc;
unsigned char hw_status;
int err;

tc = (struct tcandrv_priv *)arg;
tc->read_status_flag = 1;
tc->mhs_can_mode = MHS_CAN_HW_OFFLINE;
err = 0;
while (!kthread_should_stop()) 
  {  
  // **** Verbindungsaufbau zur Hardware
  if (tc->mhs_can_mode == MHS_CAN_HW_OFFLINE)
    {
    if ((err = TinyCanGetStatus(tc, &hw_status)) < 0)
      {  // Fehler 
      (void)schedule_timeout_interruptible(msecs_to_jiffies(500));
      continue;
      }
    else
      tc->mhs_can_mode = MHS_CAN_BUS_DO_INIT;
    continue;
    }
  // **** Tiny-CAN initialisieren
  else if (tc->mhs_can_mode == MHS_CAN_BUS_DO_INIT)
    {
    if ((err = InitTinyCan(tc)) < 0)
      tc->mhs_can_mode = MHS_CAN_HW_OFFLINE;
    else
      tc->mhs_can_mode = MHS_CAN_BUS_DO_START;
      //tc->mhs_can_mode = MHS_CAN_BUS_INIT; // <*> alt      
    continue;
    }
  // **** Bus starten
  if (tc->mhs_can_mode == MHS_CAN_BUS_DO_START)
    {
    if ((err = StartStopTinyCan(tc, 1)) < 0)
      {
      tc->mhs_can_mode = MHS_CAN_HW_OFFLINE;
      continue;
      }
    tc->mhs_can_mode = MHS_CAN_BUS_RUN;
    if (netif_queue_stopped(tc->dev))
      netif_wake_queue(tc->dev);
    continue;
    }
  else if (tc->mhs_can_mode == MHS_CAN_BUS_DO_STOP)
    {
    tc->mhs_can_mode = MHS_CAN_BUS_STOP;    
    netif_stop_queue(tc->dev);
    if ((err = StartStopTinyCan(tc, 0)) < 0)
      {
      tc->mhs_can_mode = MHS_CAN_HW_OFFLINE;
      continue;    
      }
    }
  if (tc->mhs_can_mode == MHS_CAN_BUS_RUN)
    {
    if ((err = MainCommProc(tc)) < 0)
      {
      tc->mhs_can_mode = MHS_CAN_HW_OFFLINE;
      continue;
      } 
    } 
  else
    {
    if ((err = TinyCanGetStatus(tc, &hw_status)) < 0)
      // Fehler 
      tc->mhs_can_mode = MHS_CAN_HW_OFFLINE;      
    else
      {
      if ((hw_status & 0x07) == 0x07)  // Hardware PowerUp ?
        tc->mhs_can_mode = MHS_CAN_BUS_DO_INIT;                
      } 
    (void)schedule_timeout_interruptible(msecs_to_jiffies(500));
    } 
  }
//__set_current_state(TASK_RUNNING);
//complete_and_exit(NULL, 0);
return(err);
}


static int mhs_usb_set_mode(struct net_device *netdev, enum can_mode mode)
{
struct tcandrv_priv *tc;

tc = netdev_priv(netdev);
switch (mode) 
  {
  case CAN_MODE_START: { 
                       if (tc->mhs_can_mode >= MHS_CAN_BUS_INIT)
                         tc->mhs_can_mode = MHS_CAN_BUS_DO_START;                     
                       break;
                       }
  default: return(-EOPNOTSUPP);
  }
return(0);
}


static int mhs_usb_set_bittiming(struct net_device *netdev)
{
return(0);
}


static int mhs_usb_get_berr_counter(const struct net_device *netdev,
				     struct can_berr_counter *bec)
{
struct tcandrv_priv *tc;

tc = netdev_priv(netdev);

bec->txerr = tc->bec.txerr;
bec->rxerr = tc->bec.rxerr;
return(0);
}


/******************************************
 *   Routines looking at netdevice side.
 ******************************************/

/* Send a can_frame to a TTY queue. */
static netdev_tx_t mhs_net_xmit(struct sk_buff *skb, struct net_device *dev)
{
struct tcandrv_priv *tc;
struct can_frame *from, *to;

tc = netdev_priv(dev);
if ((!tc->tty) || (skb->len != sizeof(struct can_frame)) || (!netif_running(dev)))
  {
  netdev_err(dev, "%s: xmit: iface error\n", dev->name);
  kfree_skb(skb);
  return(NETDEV_TX_OK); 
  }
spin_lock(&tc->lock);
if (!tc->tcan_thread)
  {
  spin_unlock(&tc->lock);
  kfree_skb(skb);
  return(NETDEV_TX_OK);  
  }
spin_unlock(&tc->lock);

spin_lock(&tc->fifo_lock);
if ((tc->txfifo.status & CANMSG_BUF_FULL)) 
  {
  tc->txfifo.status |= CANMSG_BUF_OVR;
  netdev_warn(dev, "CAN Tx-fifo overrun\n");
  }
else
  {
  from = (struct can_frame *) skb->data;
  to = &tc->txfifo.msg[tc->txfifo.tail];
  tc->txfifo.status |= CANMSG_BUF_PEND;
  memcpy (to, from, sizeof(struct can_frame));
  tc->txfifo.tail = (++tc->txfifo.tail) % (CANMSG_BUF_SIZE);
  if (tc->txfifo.tail == tc->txfifo.head) 
    {
    tc->txfifo.status |= CANMSG_BUF_FULL;    
    netif_stop_queue(tc->dev); // stop writing messages into the tx-fifo
    }  
  }
spin_unlock(&tc->fifo_lock);
  
kfree_skb(skb);
return(NETDEV_TX_OK);
}

  
/* Netdevice UP -> DOWN routine */
static int mhs_net_close(struct net_device *dev)
{
struct tcandrv_priv *tc;
struct task_struct *th;

tc = netdev_priv(dev);
netif_stop_queue(dev);
spin_lock(&tc->lock);
if (tc->tcan_thread)
  {
  th = tc->tcan_thread;
  tc->tcan_thread = NULL;
  spin_unlock(&tc->lock);
  kthread_stop(th);  
  }
else
  spin_unlock(&tc->lock);  
return 0;
}


/* Netdevice DOWN -> UP routine */
static int mhs_net_open(struct net_device *dev)
{
struct tcandrv_priv *tc;
int err;

err = 0;
tc = netdev_priv(dev);
if (!tc->tty)
  return(-ENODEV);
  
tc->tcan_thread = kthread_create(tcan_thread_exec, tc, DRV_NAME);
if (IS_ERR(tc->tcan_thread))
  {
  err = PTR_ERR(tc->tcan_thread);
  tc->tcan_thread = NULL;
  }
else  
  {
  if (!(err = open_candev(dev)))
    {
    tc->can.state = CAN_STATE_ERROR_ACTIVE;
    wake_up_process(tc->tcan_thread);
    netif_start_queue(dev);
    }
  }
return(err);
}


static const struct net_device_ops mhs_netdev_ops = 
  {
  .ndo_open               = mhs_net_open,
  .ndo_stop               = mhs_net_close,
  .ndo_start_xmit         = mhs_net_xmit,
  };


/******************************************
  Routines looking at TTY side.
 ******************************************/
static void setup_tty(struct tty_struct *tty, unsigned long speed)
{	
struct ktermios termios;

//termios = tty->termios;
memcpy(&termios, &tty->termios, sizeof(struct ktermios));
termios.c_iflag &=
	~(IGNBRK	// disable ignore break
	| BRKINT	// disable break causes interrupt
	| PARMRK	// disable mark parity errors
	| ISTRIP	// disable clear high bit of input characters
	| INLCR		// disable translate NL to CR 
	| IGNCR		// disable ignore CR 
	| ICRNL		// disable translate CR to NL
	| IXON);	// disable enable XON/XOFF flow control

// disable postprocess output characters
termios.c_oflag &= ~OPOST;

termios.c_lflag &=
	~(ECHO		// disable echo input characters
	| ECHONL	// disable echo new line 
	| ICANON	// disable erase, kill, werase, and rprnt
			//   special characters 
	| ISIG		// disable interrupt, quit, and suspend special
			//   characters 
	| IEXTEN);	// disable non-POSIX special characters

termios.c_cflag &=
	~(CSIZE		// no size 
	| PARENB	// disable parity bit
	| CBAUD		// clear current baud rate
	| CBAUDEX);	// clear current buad rate

termios.c_cflag |= CS8;	// character size 8 bits*/

tty_termios_encode_baud_rate(&termios, speed, speed);
tty_set_termios(tty, &termios);
}


static int mhs_tty_open(struct tty_struct *tty)
{
struct net_device *dev;
struct tcandrv_priv *priv;
struct usb_serial_port *port;
struct usb_device *udev;
int err;

if (!capable(CAP_NET_ADMIN))
  return(-EPERM);
if (!tty->ops->write)
  return(-EOPNOTSUPP);
priv = tty->disc_data;
// First make sure we're not already connected.
if ((priv) && (priv->magic == TCAN_MAGIC))
  return(-EEXIST);

if (!(dev = alloc_candev(sizeof(*priv), TX_ECHO_SKB_MAX)))
  return(-ENFILE);
priv = netdev_priv(dev);
priv->dev = dev;

udev = NULL;
if ((tty->driver) && (!strcmp(tty->driver->name, "ttyUSB")))
  {
  if ((port = (struct usb_serial_port *)(tty->driver_data)))    
    udev = (struct usb_device *)port->serial->dev;
  }

priv->tiny_can_hardware = HW_TCAN_ERROR;
if ((udev) && (udev->product))
  {
  if (!strcasecmp(udev->product, "TINY-CAN I-XL"))
    priv->tiny_can_hardware = HW_TCAN_I_XL;
  else if (!strcasecmp(udev->product, "TINY-CAN II-XL"))
    priv->tiny_can_hardware = HW_TCAN_II_XL;
  else if (!strcasecmp(udev->product, "TINY-CAN IV-XL"))
    priv->tiny_can_hardware = HW_TCAN_IV_XL;
  }

priv->can.state = CAN_STATE_STOPPED;
priv->can.clock.freq = MHS_CAN_CORE_CLOCK;
priv->can.bittiming_const = &mhs_usb_bittiming_const;
priv->can.do_set_bittiming = mhs_usb_set_bittiming;
priv->can.do_set_mode = mhs_usb_set_mode;
priv->can.do_get_berr_counter = mhs_usb_get_berr_counter;
priv->can.ctrlmode_supported = CAN_CTRLMODE_LOOPBACK | CAN_CTRLMODE_LISTENONLY;
if (priv->tiny_can_hardware == HW_TCAN_IV_XL)
  {
  priv->can.ctrlmode_supported |= (CAN_CTRLMODE_ONE_SHOT | CAN_CTRLMODE_BERR_REPORTING);
  setup_tty(tty, 8000000);
  }
else
  setup_tty(tty, 3000000);

priv->magic = TCAN_MAGIC;
priv->dev = dev;
spin_lock_init(&priv->lock);
spin_lock_init(&priv->fifo_lock);

priv->tty = tty;
tty->disc_data = priv;
dev->netdev_ops = &mhs_netdev_ops;
dev->flags |= IFF_ECHO; /* we support local echo */

if ((err = register_candev(dev)))
  {
  tty->disc_data = NULL;
  free_candev(dev);
  return(err);
  }

/* Done.  We have linked the TTY line to a channel. */
tty->receive_room = 4096; 
/* TTY layer expects 0 on success */
return(0);
}


/*
 * Close down a CAN channel.
 */
static void mhs_tty_close(struct tty_struct *tty)
{
struct tcandrv_priv *tc;
struct task_struct *th;

tc = (struct tcandrv_priv *) tty->disc_data;
/* First make sure we're connected. */
if ((!tc) || (tc->magic != TCAN_MAGIC) || (tc->tty != tty))
  return;
spin_lock(&tc->lock);
if (tc->tcan_thread)
  {
  th = tc->tcan_thread;
  tc->tcan_thread = NULL;
  spin_unlock(&tc->lock);
  kthread_stop(th);  
  }
else
  spin_unlock(&tc->lock);
tty->disc_data = NULL;
tc->tty = NULL;
unregister_netdev(tc->dev);
free_candev(tc->dev);
}


static int mhs_tty_hangup(struct tty_struct *tty)
{
mhs_tty_close(tty);
return(0);
}


/* Perform I/O control on an active TCAN channel. */
static int mhs_tty_ioctl(struct tty_struct *tty, struct file *file,
		       unsigned int cmd, unsigned long arg)
{
struct tcandrv_priv *tc;
unsigned int tmp;

tc = (struct tcandrv_priv *) tty->disc_data;
/* First make sure we're connected. */
if ((!tc) || (tc->magic != TCAN_MAGIC))
  return(-EINVAL);

switch (cmd) 
  {
  case SIOCGIFNAME   : {
                       tmp = strlen(tc->dev->name) + 1;
                       if (copy_to_user((void __user *)arg, tc->dev->name, tmp))
                         return(-EFAULT);
                       return 0;
                       }
  case SIOCSIFHWADDR : return(-EINVAL);
  default :return(tty_mode_ioctl(tty, file, cmd, arg));
  }
}


/*
 * Handle the 'receiver data ready' interrupt.
 */
static void mhs_tty_receive_buf(struct tty_struct *tty,
			      const unsigned char *cp, char *fp, int count)
{
struct tcandrv_priv *tc;
unsigned char c;

tc = (struct tcandrv_priv *)tty->disc_data;
if ((!tc) || (tc->magic != TCAN_MAGIC))
  return;
spin_lock(&tc->lock);
if (!tc->tcan_thread)
  {
  spin_unlock(&tc->lock);  
  return;  
  }
spin_unlock(&tc->lock);
while (count--) 
  {
  if (fp && *fp++) 
    {
    // Fehler !	
    cp++;
    continue;
    }
  c = *cp++;
  switch (tc->rx_state)
    {
    case RX_LOCK : break;
    case RX_WAIT : { // **** ACK empfangen                   
                   tc->ack_byte = c;
                   tc->rx_crc = CRCTabelle[c];
                   tc->rx_error = 0;
                   if ((c & 0xC0) == 0x40)          // noch Parameter Empf. ? (ACK Bit 7 = 0 Bit 6 = 1)
                     {                              // Ja =>
                     tc->rx_idx = 0;
                     tc->rx_param_cnt = ((uint32_t)c << 8) & 0x3F00;   // Anzahl Parameter (High Byte)
                     tc->rx_state = RX_PCNT;
                     }
                   else
                     {
                     if ((c & 0xE0) == 0xA0)
                       tc->rx_error = ERR_SM_NACK;                  // NACK
                     else if (c != 0x95)
                       tc->rx_error = ERR_SM_WRONG_ACK;             // Fehler beim ACK Empfang
                     tc->rx_state = RX_FIN; 
                     wake_up_process(tc->tcan_thread);
                     }
                   break;
                   }
    case RX_PCNT : {  // Anzahl Parameter (Low Byte) lesen
                   tc->rx_crc = updcrc8(c, tc->rx_crc);
                   tc->rx_param_cnt |= c;
                   tc->rx_state = RX_DATA;
                   break;
                   } 
    case RX_DATA : {
                   tc->rx_crc = updcrc8(c, tc->rx_crc);
                   tc->rx_param[tc->rx_idx++] = c;
                   if (tc->rx_idx >= TCAN_MTU)
                     {
                     tc->rx_error = ERR_PARAM_OV;
                     tc->rx_state = RX_FIN;
                     wake_up_process(tc->tcan_thread);
                     }
                   else if (tc->rx_idx == tc->rx_param_cnt)     // Alle Bytes empfangen
                     tc->rx_state = RX_CRC; 
                   break;
                   }
    case RX_CRC  : {                   
                   if (c != CRCTabelle[CRCTabelle[tc->rx_crc]]) // CRC prüfen
                     tc->rx_error = ERR_SM_WRONG_CRC;
                   tc->rx_state = RX_FIN;
                   wake_up_process(tc->tcan_thread);
                   break;
                   }
    case RX_FIN  : break;
    }
  }
}


static struct tty_ldisc_ops tcc_ldisc = {
	.owner		= THIS_MODULE,
	.magic		= TTY_LDISC_MAGIC,
	.name		= "tcan",
	.open		= mhs_tty_open,
	.close		= mhs_tty_close,
	.hangup		= mhs_tty_hangup,
	.ioctl		= mhs_tty_ioctl,
	.receive_buf	= mhs_tty_receive_buf,
};


static int __init tcan_init(void)
{
int status;

printk(banner);
/* Fill in our line protocol discipline, and register it */
if ((status = tty_register_ldisc(N_SLCAN, &tcc_ldisc)))    
  printk(KERN_ERR "tcan: can't register line discipline\n");
return status;
}


static void __exit tcan_exit(void)
{
int i;

if ((i = tty_unregister_ldisc(N_SLCAN)))
  printk(KERN_ERR "tcan: can't unregister ldisc (err %d)\n", i);
}

module_init(tcan_init);
module_exit(tcan_exit);
